#!/usr/bin/env python

#*******************************************************************************
#
#  rpnSpecial.py
#
#  rpnChilada special operators
#  copyright (c) 2021, Rick Gutleber (rickg@his.com)
#
#  License: GNU GPL 3.0 (see <http://www.gnu.org/licenses/gpl.html> for more
#  information).
#
#*******************************************************************************

import re as regex
import sys
import urllib.request as urllib2
import uuid

from urllib.error import URLError

from random import randrange

from mpmath import arange, cbrt, ceil, fabs, fadd, fdiv, fib, findpoly, \
                   floor, fmul, fsub, identify, inf, log, log10, mpmathify, \
                   phi, power, rand, root, sqrt

from rpn.rpnCombinatorics import getNthAperyNumber, getNthDelannoyNumber, getNthMenageNumber, \
                                 getNthMotzkinNumber, getNthPellNumber, getNthSchroederNumber, \
                                 getNthSchroederHipparchusNumber, getNthSylvesterNumber, getPartitionNumber

from rpn.rpnDebug import debugPrint
from rpn.rpnFactor import getFactors
from rpn.rpnGenerator import RPNGenerator

from rpn.rpnLexicographic import getErdosPersistence, getPersistence, \
                                 hasUniqueDigits, isAutomorphic, isBaseKSmithNumber, \
                                 isDecreasing, isBouncy, isIncreasing, isKaprekarNumber, isKMorphic, \
                                 isNarcissistic, isOrderKSmithNumber, isPandigital, isPerfectDigitalInvariant, \
                                 isPerfectDigitToDigitInvariant, isSmithNumber, isStepNumber, isTrimorphic, \
                                 multiplyDigits, multiplyNonzeroDigits, sumDigits

from rpn.rpnMath import isEven, isInteger, isKthPower, isOdd
from rpn.rpnName import getNumberName, getShortOrdinalName

from rpn.rpnNumberTheory import getDigitalRoot, getDivisorCount, getNthDoubleFactorial, getEulerPhi, \
                                getNthMobiusNumber, getNthAlternatingFactorial, getNthBaseKRepunit, \
                                getNthCalkinWilf, getNthCarolNumber, getNthFactorial, getNthFibonorial, \
                                getNthHyperfactorial, getNthJacobsthalNumber, getNthKFibonacciNumber, \
                                getNthKyneaNumber, getNthLeonardoNumber, getNthLucasNumber, \
                                getNthMersenneExponent, getNthMersennePrime, getNthPadovanNumber, \
                                getNthPerfectNumber, getNthSubfactorial, getNthSternNumber, \
                                getNthSuperfactorial, getNthThabitNumber, getRadical, getSigma, \
                                isAbundant, isAchillesNumber, isAntiharmonic, isCarmichaelNumber, \
                                isDeficient, isKHyperperfect, isPernicious, isPolydivisible, isPowerful, \
                                isPronic, isRough, isRuthAaronNumber, isSemiprime, isSmooth, isSphenic, \
                                isSquareFree, isUnusual

from rpn.rpnPersistence import cachedOEISFunction

from rpn.rpnPolytope import findCenteredPolygonalNumber, findPolygonalNumber, \
                            getNthCenteredPolygonalNumber, getNthPolygonalNumber

from rpn.rpnPrimeUtils import getPrimes, isPrime
from rpn.rpnUtils import oneArgFunctionEvaluator, twoArgFunctionEvaluator
from rpn.rpnValidator import argValidator, ComplexValidator, DefaultValidator, IntValidator

LARGEST_NUMBER_TO_FACTOR = power( 10, 40 )


#*******************************************************************************
#
#  getRandomNumber
#
#*******************************************************************************

def getRandomNumber( ):
    return rand( )


def getRandomNumberOperator( ):
    return getRandomNumber( )


#*******************************************************************************
#
#  getRandomInteger
#
#*******************************************************************************

def getRandomInteger( n ):
    return randrange( n )


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 0 ) ] )
def getRandomIntegerOperator( n ):
    return getRandomInteger( n )


#*******************************************************************************
#
#  getMultipleRandoms
#
#*******************************************************************************

def getMultipleRandoms( n ):
    '''
    Returns n random numbers.
    '''
    for _ in arange( 0, n ):
        yield rand( )


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 0 ) ] )
def getMultipleRandomsOperator( n ):
    return RPNGenerator.createGenerator( getMultipleRandoms, n )


#*******************************************************************************
#
#  getRandomIntegers
#
#*******************************************************************************

def getRandomIntegers( n, k ):
    '''
    Returns k random integers between 0 and n-1.
    '''
    for _ in arange( 0, k ):
        yield randrange( n )


@twoArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ), IntValidator( 0 ) ] )
def getRandomIntegersOperator( n, k ):
    return RPNGenerator.createGenerator( getRandomIntegers, [ n, k ] )


#*******************************************************************************
#
#  removeUnderscores
#
#*******************************************************************************

def removeUnderscores( source ):
    '''
    This function replaces the underscores in a string with spaces, and is
    used for formatting unit output.
    '''
    result = ''

    for c in source:
        if c == '_':
            result += ' '
        else:
            result += c

    return result


#*******************************************************************************
#
#  downloadOEISSequenceOperator
#
#*******************************************************************************

@cachedOEISFunction( 'oeis' )
def downloadOEISSequence( aNumber ):
    '''Downloads and formats data from oeis.org.'''
    keywords = downloadOEISText( aNumber, 'K' ).split( ',' )

    # If oeis.org isn't available, just punt everything
    if keywords == [ '' ]:
        return 0

    result, success = downloadOEISTable( aNumber )

    if success:
        return result

    if 'nonn' in keywords:
        result = downloadOEISText( aNumber, 'S' )
        result += downloadOEISText( aNumber, 'T' )
        result += downloadOEISText( aNumber, 'U' )
    else:
        result = downloadOEISText( aNumber, 'V' )
        result += downloadOEISText( aNumber, 'W' )
        result += downloadOEISText( aNumber, 'X' )

    if 'cons' in keywords:
        offset = int( downloadOEISText( aNumber, 'O' ).split( ',' )[ 0 ] )
        result = ''.join( result.split( ',' ) )
        return mpmathify( result[ : offset ] + '.' + result[ offset : ] )

    # return [ mpmathify( i ) for i in result.split( ',' ) ]
    return [ int( i ) for i in result.split( ',' ) ]


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ) ] )
def downloadOEISSequenceOperator( aNumber ):
    return downloadOEISSequence( aNumber )


#*******************************************************************************
#
#  downloadOEISText
#
#*******************************************************************************

def downloadOEISText( aNumber, char, addCR = False ):
    '''
    Downloads, formats and caches text data from oeis.org.
    '''
    try:
        data = urllib2.urlopen( 'http://oeis.org/search?q=id%3AA{:06}'.format( int( aNumber ) ) + '&fmt=text' ).read( )
    except URLError:
        print( 'rpn:  HTTP access to oeis.org failed' )
        return ''

    pattern = regex.compile( b'%' + bytes( char, 'ascii' ) + b' A[0-9][0-9][0-9][0-9][0-9][0-9] (.*?)\n', regex.DOTALL )

    lines = pattern.findall( data )

    result = ''

    for line in lines:
        if result != '' and addCR:
            result += '\n'

        result += line.decode( 'ascii' )

    return result


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ) ] )
def downloadOEISCommentOperator( n ):
    return downloadOEISText( n, 'C', True )


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ) ] )
def downloadOEISExtraOperator( n ):
    return downloadOEISText( n, 'E', True )


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ) ] )
def downloadOEISNameOperator( n ):
    return downloadOEISText( n, 'N', True )


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ) ] )
def downloadOEISOffsetOperator( n ):
    return int( downloadOEISText( n, 'O' ).split( ',' )[ 0 ] )


#*******************************************************************************
#
#  downloadOEISTable
#
#*******************************************************************************

def downloadOEISTable( aNumber ):
    try:
        data = urllib2.urlopen( 'http://oeis.org/A{:06}/b{:06}.txt'.format( int( aNumber ), int( aNumber ) ) ).read( )
    except URLError:
        print( 'HTTP access to oeis.org failed', file=sys.stderr )
        return [ ], False

    pattern = regex.compile( b'(.*?)[\n]', regex.DOTALL )
    lines = pattern.findall( data )

    # check if the last line doesn't end with a linefeed
    if data[ -1 ] != b'\n':
        lines.append( data[ data.rfind( b'\n' ) + 1 : ] )

    result = [ ]

    for line in lines:
        line = line.decode( 'ascii' ).strip( )

        if line == '':
            continue

        if line[ 0 ] == '#' or len( line ) == 0:
            continue

        result.append( int( line.split( )[ 1 ] ) )

    return result, True


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ) ] )
def downloadOEISTableOperator( aNumber ):
    downloadOEISTable( aNumber )


#*******************************************************************************
#
#  handleIdentify
#
#*******************************************************************************

def handleIdentify( result, file=sys.stdout ):
    '''Calls the mpmath identify function to try to identify a constant.'''
    if isinstance( result, ( list, RPNGenerator ) ):
        for i in list( result ):
            handleIdentify( i )

        return

    formula = identify( result )

    if formula is None:
        base = [ 'pi', 'e', 'euler', 'sqrt(pi)', 'phi' ]
        formula = identify( result, base )

    if formula is None:
        print( '    = [formula cannot be found]', file=file )
    else:
        print( '    = ' + formula, file=file )


#*******************************************************************************
#
#  findPolynomial
#
#*******************************************************************************

@twoArgFunctionEvaluator( )
@argValidator( [ ComplexValidator( ), IntValidator( 1 ) ] )
def findPolynomialOperator( n, k ):
    '''
    Calls the mpmath findpoly function to try to identify a polynomial of
    degree <= k for which n is a zero.
    '''
    poly = findpoly( n, int( k ) )

    if poly is None:
        poly = findpoly( n, int( k ), tol = 1e-10 )

    if poly is None:
        poly = findpoly( n, int( k ), tol = 1e-7 )

    if poly is None:
        return [ 0 ]
    else:
        return poly


#*******************************************************************************
#
#  generateUUIDOperator
#
#*******************************************************************************

def generateUUIDOperator( ):
    '''
    Generates a UUID that uses the current machine's MAC address and the
    current time as seeds.
    '''
    return str( uuid.uuid1( ) )


#*******************************************************************************
#
#  generateRandomUUIDOperator
#
#*******************************************************************************

def generateRandomUUIDOperator( ):
    '''Generates a completely random UUID.'''
    return str( uuid.uuid4( ) )


#*******************************************************************************
#
#  findInput
#
#*******************************************************************************

def findInput( value, func, estimator, minimum=0, maximum=inf ):
    guess1 = floor( estimator( value ) )

    if guess1 > maximum:
        guess1 = maximum

    if guess1 < minimum:
        guess1 = minimum

    # closing in
    result = func( guess1 )
    debugPrint( 'findInput func call', guess1 )

    if result == value:
        return True, guess1
    elif result > value:
        over = True
        delta = -1
    else:
        over = False
        delta = 1

    guess2 = guess1 + delta

    if guess2 > maximum:
        guess2 = maximum

    if guess2 < minimum:
        guess2 = minimum

    result = func( guess2 )
    debugPrint( 'findInput func call', guess2 )

    if result == value:
        return True, guess2

    if over:
        def comparator( a, b ):
            return a > b
    else:
        def comparator( a, b ):
            return a < b

    while comparator( result, value ):
        delta *= 2
        guess1 = guess2
        guess2 += delta

        if guess2 > maximum:
            return False, 0

        if guess2 < minimum:
            break

        result = func( guess2 )
        debugPrint( 'findInput func call', guess2 )

        if result == value:
            return True, guess2

    if guess1 < minimum:
        guess1 = minimum

    if guess2 < minimum:
        guess2 = minimum

    if guess1 == guess2:
        return False, 0
    elif guess1 > guess2:
        guess1, guess2 = int( guess2 ), int( guess1 )
    else:
        guess1, guess2 = int( guess1 ), int( guess2 )

    debugPrint( '->guesses:', guess1, guess2 )

    while guess1 + 1 != guess2:
        newGuess = guess1 + ( guess2 - guess1 ) // 2

        result = func( newGuess )
        debugPrint( 'findInput func call', newGuess )

        if result == value:
            return True, newGuess
        elif result > value:
            guess2 = newGuess
        else:
            guess1 = newGuess

    debugPrint( 'didn\'t find anything' )

    return False, 0


#*******************************************************************************
#
#  describeInteger
#
#*******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( 1 ) ] )
def describeIntegerOperator( n ):
    indent = ' ' * 4

    print( )
    print( int( n ), 'is:' )

    if isOdd( n ):
        print( indent + 'odd' )
    elif isEven( n ):
        print( indent + 'even' )

    if isPrime( n ):
        isPrimeFlag = True
        print( indent + 'prime' )
    elif n > 3:
        isPrimeFlag = False
        print( indent + 'composite' )
    else:
        isPrimeFlag = False

    if isKthPower( n, 2 ):
        print( indent + 'the ' + getShortOrdinalName( sqrt( n ) ) + ' square number' )

    if isKthPower( n, 3 ):
        print( indent + 'the ' + getShortOrdinalName( cbrt( n ) ) + ' cube number' )

    for i in arange( 4, fadd( ceil( log( fabs( n ), 2 ) ), 1 ) ):
        if isKthPower( n, i ):
            print( indent + 'the ' + getShortOrdinalName( root( n, i ) ) + ' ' +
                   getNumberName( i, True ) + ' power'  )

    # triangular
    guess = findPolygonalNumber( n, 3 )

    if getNthPolygonalNumber( guess, 3 ) == n:
        print( indent + 'the ' + getShortOrdinalName( guess ) + ' triangular number' )

    # pentagonal
    guess = findPolygonalNumber( n, 5 )

    if getNthPolygonalNumber( guess, 5 ) == n:
        print( indent + 'the ' + getShortOrdinalName( guess ) + ' pentagonal number' )

    # hexagonal
    guess = findPolygonalNumber( n, 6 )

    if getNthPolygonalNumber( guess, 6 ) == n:
        print( indent + 'the ' + getShortOrdinalName( guess ) + ' hexagonal number' )

    # heptagonal
    guess = findPolygonalNumber( n, 7 )

    if getNthPolygonalNumber( guess, 7 ) == n:
        print( indent + 'the ' + getShortOrdinalName( guess ) + ' heptagonal number' )

    # octagonal
    guess = findPolygonalNumber( n, 8 )

    if getNthPolygonalNumber( guess, 8 ) == n:
        print( indent + 'the ' + getShortOrdinalName( guess ) + ' octagonal number' )

    # nonagonal
    guess = findPolygonalNumber( n, 9 )

    if getNthPolygonalNumber( guess, 9 ) == n:
        print( indent + 'the ' + getShortOrdinalName( guess ) + ' nonagonal number' )

    # decagonal
    guess = findPolygonalNumber( n, 10 )

    if getNthPolygonalNumber( guess, 10 ) == n:
        print( indent + 'the ' + getShortOrdinalName( guess ) + ' decagonal number' )

    # if n > 1:
    #    for i in range( 11, 101 ):
    #        if getNthPolygonalNumber( findPolygonalNumber( n, i ), i ) == n:
    #            print( indent + str( i ) + '-gonal' )

    # centered triangular
    guess = findCenteredPolygonalNumber( n, 3 )

    if getNthCenteredPolygonalNumber( guess, 3 ) == n:
        print( indent + 'the ' + getShortOrdinalName( guess ) + ' centered triangular' )

    # centered square
    guess = findCenteredPolygonalNumber( n, 4 )

    if getNthCenteredPolygonalNumber( guess, 4 ) == n:
        print( indent + 'the ' + getShortOrdinalName( guess ) + ' centered square number' )

    # centered pentagonal
    guess = findCenteredPolygonalNumber( n, 5 )

    if getNthCenteredPolygonalNumber( guess, 5 ) == n:
        print( indent + 'the ' + getShortOrdinalName( guess ) + ' centered pentagonal number' )

    # centered hexagonal
    guess = findCenteredPolygonalNumber( n, 6 )

    if getNthCenteredPolygonalNumber( guess, 6 ) == n:
        print( indent + 'the ' + getShortOrdinalName( guess ) + ' centered hexagonal number' )

    # centered heptagonal
    guess = findCenteredPolygonalNumber( n, 7 )

    if getNthCenteredPolygonalNumber( guess, 7 ) == n:
        print( indent + 'the ' + getShortOrdinalName( guess ) + ' centered heptagonal number' )

    # centered octagonal
    guess = findCenteredPolygonalNumber( n, 8 )

    if getNthCenteredPolygonalNumber( guess, 8 ) == n:
        print( indent + 'the ' + getShortOrdinalName( guess ) + ' centered octagonal number' )

    # centered nonagonal
    guess = findCenteredPolygonalNumber( n, 9 )

    if getNthCenteredPolygonalNumber( guess, 9 ) == n:
        print( indent + 'the ' + getShortOrdinalName( guess ) + ' centered nonagonal number' )

    # centered decagonal
    guess = findCenteredPolygonalNumber( n, 10 )

    if getNthCenteredPolygonalNumber( guess, 10 ) == n:
        print( indent + 'the ' + getShortOrdinalName( guess ) + ' centered decagonal number' )

    # pandigital
    if isPandigital( n ):
        print( indent + 'pandigital' )

    # for i in range( 4, 21 ):
    #    if isBaseKPandigital( n, i ):
    #        print( indent + 'base ' + str( i ) + ' pandigital' )

    # Fibonacci
    result = findInput( n, fib, lambda n: fmul( log10( n ), 5 ) )

    if result[ 0 ]:
        print( indent + 'the ' + getShortOrdinalName( result[ 1 ] ) + ' Fibonacci number' )

    # Tribonacci
    result = findInput( n, lambda n : getNthKFibonacciNumber( n, 3 ), lambda n: fmul( log10( n ), 5 ) )

    if result[ 0 ]:
        print( indent + 'the ' + getShortOrdinalName( result[ 1 ] ) + ' Tribonacci number' )

    # Tetranacci
    result = findInput( n, lambda n : getNthKFibonacciNumber( n, 4 ), lambda n: fmul( log10( n ), 5 ) )

    if result[ 0 ]:
        print( indent + 'the ' + getShortOrdinalName( result[ 1 ] ) + ' Tetranacci number' )

    # Pentanacci
    result = findInput( n, lambda n : getNthKFibonacciNumber( n, 5 ), lambda n: fmul( log10( n ), 5 ) )

    if result[ 0 ]:
        print( indent + 'the ' + getShortOrdinalName( result[ 1 ] ) + ' Pentanacci number' )

    # Hexanacci
    result = findInput( n, lambda n : getNthKFibonacciNumber( n, 6 ), lambda n: fmul( log10( n ), 5 ) )

    if result[ 0 ]:
        print( indent + 'the ' + getShortOrdinalName( result[ 1 ] ) + ' Hexanacci number' )

    # Heptanacci
    result = findInput( n, lambda n : getNthKFibonacciNumber( n, 7 ), lambda n: fmul( log10( n ), 5 ) )

    if result[ 0 ]:
        print( indent + 'the ' + getShortOrdinalName( result[ 1 ] ) + ' Heptanacci number' )

    # Octanacci
    result = findInput( n, lambda n : getNthKFibonacciNumber( n, 8 ), lambda n: fmul( log10( n ), 5 ) )

    if result[ 0 ]:
        print( indent + 'the ' + getShortOrdinalName( result[ 1 ] ) + ' Octanacci number' )

    # Lucas numbers
    result = findInput( n, getNthLucasNumber, lambda n: fmul( log10( n ), 5 ) )

    if result[ 0 ]:
        print( indent + 'the ' + getShortOrdinalName( result[ 1 ] ) + ' Lucas number' )

    # base-k repunits
    if n > 1:
        for i in range( 2, 21 ):
            result = findInput( n, lambda x: getNthBaseKRepunit( x, i ), lambda n: log( n, i ), minimum=1 )

            if result[ 0 ]:
                print( indent + 'the ' + getShortOrdinalName( result[ 1 ] ) + ' base-' + str( i ) + ' repunit' )

    # Jacobsthal numbers
    result = findInput( n, getNthJacobsthalNumber, lambda n: fmul( log( n ), 1.6 ) )

    if result[ 0 ]:
        print( indent + 'the ' + getShortOrdinalName( result[ 1 ] ) + ' Jacobsthal number' )

    # Padovan numbers
    result = findInput( n, getNthPadovanNumber, lambda n: fmul( log10( n ), 9 ) )

    if result[ 0 ]:
        print( indent + 'the ' + getShortOrdinalName( result[ 1 ] ) + ' Padovan number' )

    # Fibonorial numbers
    result = findInput( n, getNthFibonorial, lambda n: sqrt( fmul( log( n ), 10 ) ) )

    if result[ 0 ]:
        print( indent + 'the ' + getShortOrdinalName( result[ 1 ] ) + ' Fibonorial number' )

    # Mersenne primes
    result = findInput( n, getNthMersennePrime, lambda n: fadd( fmul( log( log( sqrt( n ) ) ), 2.7 ), 3 ) )

    if result[ 0 ]:
        print( indent + 'the ' + getShortOrdinalName( result[ 1 ] ) + ' Mersenne prime' )

    # perfect number
    result = findInput( n, getNthPerfectNumber, lambda n: fmul( log( log( n ) ), 2.6 ) )

    if result[ 0 ]:
        print( indent + 'the ' + getShortOrdinalName( result[ 1 ] ) + ' perfect number' )

    # Mersenne exponent
    result = findInput( n, getNthMersenneExponent, lambda n: fmul( log( n ), 2.7 ), maximum=50 )

    if result[ 0 ]:
        print( indent + 'the ' + getShortOrdinalName( result[ 1 ] ) + ' Mersenne exponent' )

    if not isPrimeFlag and n != 1 and n <= LARGEST_NUMBER_TO_FACTOR:
        # deficient
        if isDeficient( n ):
            print( indent + 'deficient' )

        # abundant
        if isAbundant( n ):
            print( indent + 'abundant' )

        # k_hyperperfect
        for i in sorted( list( set( sorted( downloadOEISSequence( 34898 )[ : 500 ] ) ) ) )[ 1 : ]:
            if i > n:
                break

            if isKHyperperfect( n, i ):
                print( indent + str( i ) + '-hyperperfect' )
                break

        # smooth
        for i in getPrimes( 2, 50 ):
            if isSmooth( n, i ):
                print( indent + str( i ) + '-smooth' )
                break

        # rough
        previousPrime = 2

        for i in getPrimes( 2, 50 ):
            if not isRough( n, i ):
                print( indent + str( previousPrime ) + '-rough' )
                break

            previousPrime = i

        # is_semiprime
        if isSemiprime( n ):
            print( indent + 'semiprime' )

        # is_sphenic
        elif isSphenic( n ):
            print( indent + 'sphenic' )
        elif isSquareFree( n ):
            print( indent + 'square-free' )

        # powerful
        if isPowerful( n ):
            print( indent + 'powerful' )

    # factorial
    result = findInput( n, getNthFactorial, lambda n: power( log10( n ), 0.92 ) )

    if result[ 0 ]:
        print( indent + 'the ' + getShortOrdinalName( result[ 1 ] ) + ' factorial number' )

    # alternating factorial
    result = findInput( n, getNthAlternatingFactorial, lambda n: fmul( sqrt( log( n ) ), 0.72 ) )

    if result[ 0 ]:
        print( indent + 'the ' + getShortOrdinalName( result[ 1 ] ) + ' alternating factorial number' )

    # double factorial
    if n == 1:
        result = ( True, 1 )
    else:
        result = findInput( n, getNthDoubleFactorial, lambda n: fdiv( power( log( log( n ) ), 4 ), 7 ) )

    if result[ 0 ]:
        print( indent + 'the ' + getShortOrdinalName( result[ 1 ] ) + ' double factorial number' )

    # hyperfactorial
    result = findInput( n, getNthHyperfactorial, lambda n: fmul( sqrt( log( n ) ), 0.8 ) )

    if result[ 0 ]:
        print( indent + 'the ' + getShortOrdinalName( result[ 1 ] ) + ' hyperfactorial number' )

    # subfactorial
    result = findInput( n, getNthSubfactorial, lambda n: fmul( log10( n ), 1.1 ) )

    if result[ 0 ]:
        print( indent + 'the ' + getShortOrdinalName( result[ 1 ] ) + ' subfactorial number' )

    # superfactorial
    if n == 1:
        result = ( True, 1 )
    else:
        result = findInput( n, getNthSuperfactorial, lambda n: fadd( sqrt( log( n ) ), 1 ) )

    if result[ 0 ]:
        print( indent + 'the ' + getShortOrdinalName( result[ 1 ] ) + ' superfactorial number' )

    # pernicious
    if isPernicious( n ):
        print( indent + 'pernicious' )

    # pronic
    if isPronic( n ):
        print( indent + 'pronic' )

    # Achilles
    if isAchillesNumber( n ):
        print( indent + 'an Achilles number' )

    # antiharmonic
    if isAntiharmonic( n ):
        print( indent + 'antiharmonic' )

    # unusual
    if isUnusual( n ):
        print( indent + 'unusual' )

    # hyperperfect
    for i in range( 2, 21 ):
        if isKHyperperfect( n, i ):
            print( indent + str( i ) + '-hyperperfect' )

    # Ruth-Aaron
    if isRuthAaronNumber( n ):
        print( indent + 'a Ruth-Aaron number' )

    # Smith numbers
    if isSmithNumber( n ):
        print( indent + 'a Smith number' )

    # base-k Smith numbers
    for i in range( 2, 10 ):
        if isBaseKSmithNumber( n, i ):
            print( indent + 'a base-' + str( i ) + ' Smith number' )

    # order-k Smith numbers
    for i in range( 2, 11 ):
        if isOrderKSmithNumber( n, i ):
            print( indent + 'an order-' + str( i ) + ' Smith number' )

    # polydivisible
    if isPolydivisible( n ):
        print( indent + 'polydivisible' )

    # Carol numbers
    result = findInput( n, getNthCarolNumber, lambda n: fmul( log10( n ), fdiv( 5, 3 ) ) )

    if result[ 0 ]:
        print( indent + 'the ' + getShortOrdinalName( result[ 1 ] ) + ' Carol number' )

    # Kynea numbers
    result = findInput( n, getNthKyneaNumber, lambda n: fmul( log10( n ), fdiv( 5, 3 ) ) )

    if result[ 0 ]:
        print( indent + 'the ' + getShortOrdinalName( result[ 1 ] ) + ' Kynea number' )

    # Leonardo numbers
    result = findInput( n, getNthLeonardoNumber, lambda n: fsub( log( n, phi ), fdiv( 1, phi ) ) )

    if result[ 0 ]:
        print( indent + 'the ' + getShortOrdinalName( result[ 1 ] ) + ' Leonardo number' )

    # Thabit numbers
    result = findInput( n, getNthThabitNumber, lambda n: fmul( log10( n ), 3.25 ) )

    if result[ 0 ]:
        print( indent + 'the ' + getShortOrdinalName( result[ 1 ] ) + ' Thabit number' )

    # Carmmichael
    if isCarmichaelNumber( n ):
        print( indent + 'a Carmichael number' )

    # narcissistic
    if isNarcissistic( n ):
        print( indent + 'narcissistic' )

    # PDI
    if isPerfectDigitalInvariant( n ):
        print( indent + 'a perfect digital invariant' )

    # PDDI
    if isPerfectDigitToDigitInvariant( n, 10 ):
        print( indent + 'a perfect digit-to-digit invariant in base 10' )

    # Kaprekar
    if isKaprekarNumber( n ):
        print( indent + 'a Kaprekar number' )

    # automorphic
    if isAutomorphic( n ):
        print( indent + 'automorphic' )

    # trimorphic
    if isTrimorphic( n ):
        print( indent + 'trimorphic' )

    # k-morphic
    for i in range( 4, 11 ):
        if isKMorphic( n, i ):
            print( indent + str( i ) + '-morphic' )

    # bouncy
    if isBouncy( n ):
        print( indent + 'bouncy' )

    # step number
    if isStepNumber( n ):
        print( indent + 'a step number' )

    # Apery numbers
    result = findInput( n, getNthAperyNumber, lambda n: fadd( fdiv( log( n ), 1.5 ), 1 ) )

    if result[ 0 ]:
        print( indent + 'the ' + getShortOrdinalName( result[ 1 ] ) + ' Apery number' )

    # Delannoy numbers
    result = findInput( n, getNthDelannoyNumber, lambda n: fmul( log10( n ), 1.35 ) )

    if result[ 0 ]:
        print( indent + 'the ' + getShortOrdinalName( result[ 1 ] ) + ' Delannoy number' )

    # Schroeder numbers
    result = findInput( n, getNthSchroederNumber, lambda n: fmul( log10( n ), 1.6 ) )

    if result[ 0 ]:
        print( indent + 'the ' + getShortOrdinalName( result[ 1 ] ) + ' Schroeder number' )

    # Schroeder-Hipparchus numbers
    result = findInput( n, getNthSchroederHipparchusNumber, lambda n: fdiv( log10( n ), 1.5 ) )

    if result[ 0 ]:
        print( indent + 'the ' + getShortOrdinalName( result[ 1 ] ) + ' Schroeder-Hipparchus number' )

    # Motzkin numbers
    result = findInput( n, getNthMotzkinNumber, log )

    if result[ 0 ]:
        print( indent + 'the ' + getShortOrdinalName( result[ 1 ] ) + ' Motzkin number' )

    # Pell numbers
    result = findInput( n, getNthPellNumber, lambda n: fmul( log( n ), 1.2 ) )

    if result[ 0 ]:
        print( indent + 'the ' + getShortOrdinalName( result[ 1 ] ) + ' Pell number' )

    # Sylvester numbers
    if n > 1:
        result = findInput( n, getNthSylvesterNumber, lambda n: sqrt( sqrt( log( n ) ) ) )

        if result[ 0 ]:
            print( indent + 'the ' + getShortOrdinalName( result[ 1 ] ) + ' Sylvester number' )

    # partition numbers
    result = findInput( n, getPartitionNumber, lambda n: power( log( n ), 1.56 ) )

    if result[ 0 ]:
        print( indent + 'the ' + getShortOrdinalName( result[ 1 ] ) + ' partition number' )

    # menage numbers
    result = findInput( n, getNthMenageNumber, lambda n: fdiv( log10( n ), 1.2 ) )

    if result[ 0 ]:
        print( indent + 'the ' + getShortOrdinalName( result[ 1 ] ) + ' menage number' )

    print( )
    print( int( n ), 'has:' )

    # number of digits
    digits = log10( n )

    if isInteger( digits ):
        digits += 1
    else:
        digits = ceil( digits )

    print( indent + str( int( digits ) ) + ' digit' + ( 's' if digits > 1 else '' ) )

    # digit sum
    print( indent + 'a digit sum of ' + str( int( sumDigits( n ) ) ) )

    # digit product
    digitProduct = multiplyDigits( n )

    print( indent + 'a digit product of ' + str( int( digitProduct ) ) )

    # non-zero digit product
    if digitProduct == 0:
        print( indent + 'a non-zero digit product of ' + str( int( multiplyNonzeroDigits( n ) ) ) )

    if not isPrimeFlag and n != 1 and n <= LARGEST_NUMBER_TO_FACTOR:
        # factors
        factors = getFactors( n )
        factorCount = len( factors )
        print( indent + str( factorCount ) + ' prime factor' + ( 's' if factorCount > 1 else '' ) +
               ': ' + ', '.join( [ str( int( i ) ) for i in factors ] ) )

        # number of divisors
        divisorCount = int( getDivisorCount( n ) )
        print( indent + str( divisorCount ) + ' divisor' + ( 's' if divisorCount > 1 else '' ) )

    if n <= LARGEST_NUMBER_TO_FACTOR:
        print( indent + 'a divisor sum of ' + str( int( getSigma( n ) ) ) )
        print( indent + 'a Stern value of ' + str( int( getNthSternNumber( n ) ) ) )
        calkinWilf = getNthCalkinWilf( n )
        print( indent + 'a Calkin-Wilf value of ' + str( int( calkinWilf[ 0 ] ) ) + '/' +
               str( int( calkinWilf[ 1 ] ) ) )
        print( indent + 'a Mobius value of ' + str( int( getNthMobiusNumber( n ) ) ) )
        print( indent + 'a radical of ' + str( int( getRadical( n ) ) ) )
        print( indent + 'a Euler phi value of ' + str( int( getEulerPhi( n ) ) ) )

    print( indent + 'a digital root of ' + str( int( getDigitalRoot( n ) ) ) )

    if hasUniqueDigits( n ):
        print( indent + 'unique digits' )

    print( indent + 'a multiplicative persistence of ' + str( int( getPersistence( n ) ) ) )
    print( indent + 'an Erdos persistence of ' + str( int( getErdosPersistence( n ) ) ) )

    if n > 9 and isIncreasing( n ):
        if not isDecreasing( n ):
            print( indent + 'increasing digits' )
    elif n > 9 and isDecreasing( n ):
        print( indent + 'decreasing digits' )

    print( )

    return n


#*******************************************************************************
#
#  ifOperator
#
#*******************************************************************************

@argValidator( [ DefaultValidator( ), DefaultValidator( ), ComplexValidator( ) ] )
def ifOperator( a, b, c ):
    return a if c else b



#!/usr/bin/env python

#******************************************************************************
#
#  rpnMath.py
#
#  rpnChilada mathematical operators
#  copyright (c) 2021, Rick Gutleber (rickg@his.com)
#
#  License: GNU GPL 3.0 (see <http://www.gnu.org/licenses/gpl.html> for more
#  information).
#
#******************************************************************************

from functools import reduce

from hyperop import hyperop
from mpmath import acos, acosh, acot, acoth, acsc, acsch, agm, arange, arg, \
                   asec, asech, asin, asinh, atan, atanh, autoprec, ceil, \
                   conj, cos, cosh, cot, coth, csc, csch, e, exp, fabs, fadd, \
                   fdiv, floor, fmod, fmul, fneg, fsub, hypot, im, isint, \
                   lambertw, li, ln, log, log10, mpf, nint, phi, polyexp, \
                   polylog, power, re, root, sec, sech, sign, sin, sinh, sqrt, \
                   tan, tanh, unitroots

from rpn.rpnDateTime import RPNDateTime
from rpn.rpnGenerator import RPNGenerator
from rpn.rpnMeasurementClass import RPNMeasurement
from rpn.rpnUtils import oneArgFunctionEvaluator, twoArgFunctionEvaluator, listArgFunctionEvaluator
from rpn.rpnValidator import argValidator, AdditiveValidator, ComparableValidator, IntValidator, \
                             ListValidator, MultiplicativeValidator, RealValidator, RealOrMeasurementValidator


#******************************************************************************
#
#  addOperator
#
#  We used to be able to call fadd directly, but now we want to be able to add
#  units.  Adding units includes an implicit conversion if the units are not
#  the same, assuming they are compatible.
#
#******************************************************************************

def add( n, k ):
    if isinstance( n, RPNDateTime ) and isinstance( k, RPNMeasurement ):
        return n.add( k )

    if isinstance( n, RPNMeasurement ) and isinstance( k, RPNDateTime ):
        return k.add( n )

    if isinstance( n, RPNMeasurement ):
        return n.add( k )

    if isinstance( k, RPNMeasurement ):
        return RPNMeasurement( n ).add( k )

    return fadd( n, k )


@twoArgFunctionEvaluator( )
# note, a date-time can be added to, but it cannot itself be added
@argValidator( [ AdditiveValidator( ), MultiplicativeValidator( ) ] )
def addOperator( n, k ):
    return add( n, k )


@oneArgFunctionEvaluator( )
@argValidator( [ AdditiveValidator( ) ] )
def incrementOperator( n ):
    return add( n, 1 )


#******************************************************************************
#
#  subtractOperator
#
#  We used to be able to call fsub directly, but now we want to be able to
#  subtract units and do the appropriate conversions.
#
#******************************************************************************

def subtract( n, k ):
    if isinstance( n, RPNDateTime ):
        return n.subtract( k )

    if isinstance( n, RPNMeasurement ):
        if isinstance( k, RPNDateTime ):
            return k.subtract( n )

        return n.subtract( k )

    if isinstance( k, RPNMeasurement ):
        return RPNMeasurement( n ).subtract( k )

    return fsub( n, k )


@twoArgFunctionEvaluator( )
@argValidator( [ AdditiveValidator( ), AdditiveValidator( ) ] )
def subtractOperator( n, k ):
    return subtract( n, k )


@oneArgFunctionEvaluator( )
@argValidator( [ AdditiveValidator( ) ] )
def decrementOperator( n ):
    return subtract( n, 1 )


#******************************************************************************
#
#  getNegativeOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ) ] )
def getNegativeOperator( n ):
    if isinstance( n, RPNMeasurement ):
        return RPNMeasurement( fneg( n.value ), n.units )

    return fneg( n )


#******************************************************************************
#
#  getSignOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ) ] )
def getSignOperator( n ):
    if isinstance( n, RPNMeasurement ):
        return sign( n.value )

    return sign( n )


#******************************************************************************
#
#  getValueOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ) ] )
def getValueOperator( n ):
    if isinstance( n, RPNMeasurement ):
        return n.value

    return n


#******************************************************************************
#
#  divideOperator
#
#  We used to be able to call fdiv directly, but now we want to also divide
#  the units.  Doing so lets us do all kinds of great stuff because now we
#  can support compound units without having to explicitly declare them in
#  makeUnits.py.
#
#******************************************************************************

def divide( n, k ):
    if isinstance( n, RPNMeasurement ):
        return n.divide( k )

    if isinstance( k, RPNMeasurement ):
        return RPNMeasurement( n ).divide( k )

    return fdiv( n, k )


@twoArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ), MultiplicativeValidator( ) ] )
def divideOperator( n, k ):
    return divide( n, k )


#******************************************************************************
#
#  multiplyOperator
#
#  We used to be able to call fmul directly, but now we want to also multiply
#  the units.  This allows compound units and the conversion routines try to
#  be smart enough to deal with this.
#
#******************************************************************************

def multiply( n, k ):
    if isinstance( n, RPNMeasurement ):
        return n.multiply( k )

    if isinstance( k, RPNMeasurement ):
        return RPNMeasurement( n ).multiply( k )

    return fmul( n, k )


@twoArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ), MultiplicativeValidator( ) ] )
def multiplyOperator( n, k ):
    return multiply( n, k )


#******************************************************************************
#
#  getPowerOperator
#
#******************************************************************************

def getPower( n, k ):
    if isinstance( n, RPNMeasurement ):
        result = RPNMeasurement( n )
        return result.exponentiate( k )

    return power( n, k )


@twoArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ), MultiplicativeValidator( ) ] )
def getPowerOperator( n, k ):
    return getPower( n, k )


def square( n ):
    return getPower( n, 2 )


@oneArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ) ] )
def squareOperator( n ):
    return getPower( n, 2 )


@oneArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ) ] )
def cubeOperator( n ):
    return getPower( n, 3 )


#******************************************************************************
#
#  getRootOperator
#
#******************************************************************************

def getRoot( n, k ):
    if isinstance( n, RPNMeasurement ):
        return n.getRoot( k )

    if not isint( k ):
        return power( n, fdiv( 1, k ) )

    return root( n, k )


@twoArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ), MultiplicativeValidator( ) ] )
def getRootOperator( n, k ):
    return getRoot( n, k )


@oneArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ) ] )
def getSquareRootOperator( n ):
    return getRoot( n, 2 )


@oneArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ) ] )
def getCubeRootOperator( n ):
    return getRoot( n, 3 )


#******************************************************************************
#
#  getSquareSuperRootOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ) ] )
def getSquareSuperRootOperator( n ):
    '''Returns the positive, real square super root of n.'''
    return power( e, lambertw( log( n ) ) )


#******************************************************************************
#
#  getCubeSuperRootOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ) ] )
def getCubeSuperRootOperator( n ):
    '''Returns the positive, real cube super root of n.'''
    value = fmul( 2, log( n ) )
    return sqrt( fdiv( value, lambertw( value ) ) )


#******************************************************************************
#
#  getSuperRootOperator
#
#******************************************************************************

@twoArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ), IntValidator( 1 ) ] )
def getSuperRootOperator( n, k ):
    '''Returns the positive, real kth super root of n.'''
    k = fsub( k, 1 )
    value = fmul( k, log( n ) )
    return root( fdiv( value, lambertw( value ) ), k )


#******************************************************************************
#
#  getSuperRootsOperator
#
#******************************************************************************

@twoArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ), IntValidator( 1 ) ] )
def getSuperRootsOperator( n, k ):
    '''Returns all the super-roots of n, not just the nice, positive, real one.'''
    k = fsub( k, 1 )
    factors = [ fmul( i, root( k, k ) ) for i in unitroots( int( k ) ) ]
    base = root( fdiv( log( n ), lambertw( fmul( k, log( n ) ) ) ), k )

    return [ fmul( i, base ) for i in factors ]


#******************************************************************************
#
#  getReciprocalOperator
#
#  We used to be able to call fdiv directly, but now we want to handle
#  RPNMeasurements.
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ) ] )
def getReciprocalOperator( n ):
    if isinstance( n, RPNMeasurement ):
        return n.invert( invertValue=False )

    return fdiv( 1, n )


#******************************************************************************
#
#  getAbsoluteValueOperator
#
#  We used to be able to call fabs directly, but now we want to handle
#  RPNMeasurements.
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ) ] )
def getAbsoluteValueOperator( n ):
    if isinstance( n, RPNMeasurement ):
        return RPNMeasurement( fabs( n.value ), n.units )

    return fabs( n )


#******************************************************************************
#
#  getNearestIntOperator
#
#  We used to be able to call fdiv directly, but now we want to handle
#  RPNMeasurements.
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ) ] )
def getNearestIntOperator( n ):
    if isinstance( n, RPNMeasurement ):
        return RPNMeasurement( nint( n.value ), n.units )

    return nint( n )


#******************************************************************************
#
#  tetrateOperator
#
#******************************************************************************

@twoArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ), IntValidator( 0 ) ] )
def tetrateOperator( i, j ):
    '''
    This is the smaller (left-associative) version of the hyper4 operator.

    This function forces the second argument to an integer and runs in O( n )
    time based on the second argument.
    '''
    result = i

    for _ in arange( 1, j ):
        result = power( result, i )

    return result


#******************************************************************************
#
#  tetrateRightOperator
#
#******************************************************************************

@twoArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ), IntValidator( 0 ) ] )
def tetrateRightOperator( i, j ):
    '''
    This is the larger, right-associative version of the hyper4 operator.

    This function forces the second argument to an integer and runs in O( n )
    time based on the second argument.
    '''
    result = i

    for _ in arange( 1, j ):
        result = power( i, result )

    return result


#******************************************************************************
#
#  isDivisibleOperator
#
#  Is n divisible by k?
#
#******************************************************************************

def isDivisible( n, k ):
    if n == 0:
        return 1

    return 1 if ( n >= k ) and ( fmod( n, k ) == 0 ) else 0


@twoArgFunctionEvaluator( )
#TODO: handle measurements
@argValidator( [ MultiplicativeValidator( ), MultiplicativeValidator( ) ] )
def isDivisibleOperator( n, k ):
    return isDivisible( n, k )


#******************************************************************************
#
#  isSquareOperator
#
#  The "smarter" algorithm is slower... WHY?!
#
#******************************************************************************

#TODO: handle measurements
def isSquare( n ):
    #mod = fmod( n, 16 )

    #if mod in [ 0, 1, 4, 9 ]:
    sqrtN = sqrt( n )
    return 1 if sqrtN == floor( sqrtN ) else 0
    #else:
    #    return 0


@oneArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ) ] )
def isSquareOperator( n ):
    return isSquare( n )


#******************************************************************************
#
#  isPowerOperator
#
#******************************************************************************

@twoArgFunctionEvaluator( )
#TODO: handle measurements
@argValidator( [ MultiplicativeValidator( ), MultiplicativeValidator( ) ] )
def isPowerOperator( n, k ):
    #print( 'log( n )', log( n ) )
    #print( 'log( k )', log( k ) )
    #print( 'divide', autoprec( lambda n, k: fdiv( re( log( n ) ), re( log( k ) ) ) )( n, k ) )
    return isInteger( autoprec( lambda n, k: fdiv( re( log( n ) ), re( log( k ) ) ) )( n, k ) )


#******************************************************************************
#
#  isKthPowerOperator
#
#******************************************************************************

def isKthPower( n, k ):
    if not isint( k, gaussian=True ):
        raise ValueError( 'integer argument expected' )

    if k == 1:
        return 1

    if im( n ):
        # I'm not sure why this is necessary...
        if re( n ) == 0:
            return isKthPower( im( n ), k )

        # We're looking for a Gaussian integer among any of the roots.
        for i in [ autoprec( root )( n, k, i ) for i in arange( k ) ]:
            if isint( i, gaussian=True ):
                return 1

        return 0

    rootN = autoprec( root )( n, k )
    return 1 if isint( rootN, gaussian=True ) else 0


@twoArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ), IntValidator( 1 ) ] )
def isKthPowerOperator( n, k ):
    return isKthPower( n, k )


#******************************************************************************
#
#  performTrigOperation
#
#******************************************************************************

def performTrigOperation( i, operation ):
    if isinstance( i, RPNMeasurement ):
        value = mpf( i.convertValue( RPNMeasurement( 1, { 'radian' : 1 } ) ) )
    else:
        value = i

    return operation( value )


@oneArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ) ] )
def acosOperator( n ):
    return performTrigOperation( n, acos )


@oneArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ) ] )
def acoshOperator( n ):
    return performTrigOperation( n, acosh )


@oneArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ) ] )
def acotOperator( n ):
    return performTrigOperation( n, acot )


@oneArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ) ] )
def acothOperator( n ):
    return performTrigOperation( n, acoth )


@oneArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ) ] )
def acscOperator( n ):
    return performTrigOperation( n, acsc )


@oneArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ) ] )
def acschOperator( n ):
    return performTrigOperation( n, acsch )


@oneArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ) ] )
def asecOperator( n ):
    return performTrigOperation( n, asec )


@oneArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ) ] )
def asechOperator( n ):
    return performTrigOperation( n, asech )


@oneArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ) ] )
def asinOperator( n ):
    return performTrigOperation( n, asin )


@oneArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ) ] )
def asinhOperator( n ):
    return performTrigOperation( n, asinh )


@oneArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ) ] )
def atanOperator( n ):
    return performTrigOperation( n, atan )


@oneArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ) ] )
def atanhOperator( n ):
    return performTrigOperation( n, atanh )


@oneArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ) ] )
def cosOperator( n ):
    return performTrigOperation( n, cos )


@oneArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ) ] )
def coshOperator( n ):
    return performTrigOperation( n, cosh )


@oneArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ) ] )
def cotOperator( n ):
    return performTrigOperation( n, cot )


@oneArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ) ] )
def cothOperator( n ):
    return performTrigOperation( n, coth )


@oneArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ) ] )
def cscOperator( n ):
    return performTrigOperation( n, csc )


@oneArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ) ] )
def cschOperator( n ):
    return performTrigOperation( n, csch )


@oneArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ) ] )
def secOperator( n ):
    return performTrigOperation( n, sec )


@oneArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ) ] )
def sechOperator( n ):
    return performTrigOperation( n, sech )


@oneArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ) ] )
def sinOperator( n ):
    return performTrigOperation( n, sin )


@oneArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ) ] )
def sinhOperator( n ):
    return performTrigOperation( n, sinh )


@oneArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ) ] )
def tanOperator( n ):
    return performTrigOperation( n, tan )


@oneArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ) ] )
def tanhOperator( n ):
    return performTrigOperation( n, tanh )


#******************************************************************************
#
#  isEqualOperator
#
#******************************************************************************

def isEqual( n, k ):
    if isinstance( n, RPNMeasurement ):
        return 1 if n.isEqual( k ) else 0

    if isinstance( n, RPNMeasurement ):
        return 1 if k.isEqual( n ) else 0

    return 1 if n == k else 0


@twoArgFunctionEvaluator( )
@argValidator( [ ComparableValidator( ), ComparableValidator( ) ] )
def isEqualOperator( n, k ):
    return isEqual( n, k )


#******************************************************************************
#
#  isNotEqualOperator
#
#******************************************************************************

@twoArgFunctionEvaluator( )
@argValidator( [ ComparableValidator( ), ComparableValidator( ) ] )
def isNotEqualOperator( n, k ):
    return 0 if isEqual( n, k ) else 1


#******************************************************************************
#
#  isGreaterOperator
#
#******************************************************************************

def isGreater( n, k ):
    if isinstance( n, RPNMeasurement ):
        return 1 if n.isLarger( k ) else 0

    return 1 if n > k else 0


@twoArgFunctionEvaluator( )
@argValidator( [ ComparableValidator( ), ComparableValidator( ) ] )
def isGreaterOperator( n, k ):
    return isGreater( n, k )


#******************************************************************************
#
#  isLessOperator
#
#******************************************************************************

def isLess( n, k ):
    if isinstance( n, RPNMeasurement ):
        return 1 if n.isSmaller( k ) else 0

    return 1 if n < k else 0


@twoArgFunctionEvaluator( )
@argValidator( [ ComparableValidator( ), ComparableValidator( ) ] )
def isLessOperator( n, k ):
    return isLess( n, k )


#******************************************************************************
#
#  isNotGreaterOperator
#
#******************************************************************************

@twoArgFunctionEvaluator( )
@argValidator( [ ComparableValidator( ), ComparableValidator( ) ] )
def isNotGreaterOperator( n, k ):
    if isinstance( n, RPNMeasurement ):
        return 1 if n.isNotLarger( k ) else 0

    return 1 if n <= k else 0


#******************************************************************************
#
#  isNotLessOperator
#
#******************************************************************************

@twoArgFunctionEvaluator( )
@argValidator( [ ComparableValidator( ), ComparableValidator( ) ] )
def isNotLessOperator( n, k ):
    if isinstance( n, RPNMeasurement ):
        return 1 if n.isNotSmaller( k ) else 0

    return 1 if n >= k else 0


#******************************************************************************
#
#  isIntegerOperator
#
#******************************************************************************

def isInteger( n ):
    return 1 if fmod( n, 1 ) == 0 else 0


@oneArgFunctionEvaluator( )
@argValidator( [ RealOrMeasurementValidator( ) ] )
def isIntegerOperator( n ):
    return isInteger( n )


#******************************************************************************
#
#  roundOffOperator
#
#******************************************************************************

def roundOff( n ):
    if isinstance( n, RPNMeasurement ):
        return RPNMeasurement( roundOff( n.value ), n.units )

    return floor( fadd( n, 0.5 ) )


@oneArgFunctionEvaluator( )
@argValidator( [ RealOrMeasurementValidator( ) ] )
def roundOffOperator( n ):
    return roundOff( n )


#******************************************************************************
#
#  roundByValueOperator
#
#******************************************************************************

def roundByValue( n, value ):
    if isinstance( n, RPNMeasurement ):
        return RPNMeasurement( roundByValue( n.value, value ), n.units )

    return fmul( floor( fdiv( fadd( n, fdiv( value, 2 ) ), value ) ), value )


@twoArgFunctionEvaluator( )
@argValidator( [ RealOrMeasurementValidator( ), RealValidator( ) ] )
def roundByValueOperator( n, value ):
    return roundByValue( n, value )


#******************************************************************************
#
#  roundByDigitsOperator
#
#******************************************************************************

def roundByDigits( n, digits ):
    if isinstance( n, RPNMeasurement ):
        return RPNMeasurement( roundByDigits( n.value, digits ), n.units )

    return roundByValue( n, power( 10, digits ) )


@twoArgFunctionEvaluator( )
@argValidator( [ RealOrMeasurementValidator( ), IntValidator( ) ] )
def roundByDigitsOperator( n, digits ):
    return roundByDigits( n, digits )


#******************************************************************************
#
#  getLargerOperator
#
#******************************************************************************

@twoArgFunctionEvaluator( )
@argValidator( [ ComparableValidator( ), ComparableValidator( ) ] )
def getLargerOperator( n, k ):
    if isinstance( n, RPNMeasurement ):
        return n if isGreater( n, k ) else k

    return n if n > k else k


#******************************************************************************
#
#  getSmallerOperator
#
#******************************************************************************

@twoArgFunctionEvaluator( )
@argValidator( [ ComparableValidator( ), ComparableValidator( ) ] )
def getSmallerOperator( n, k ):
    if isinstance( n, RPNMeasurement ):
        return n if isLess( n, k ) else k

    return n if n < k else k


#******************************************************************************
#
#  getFloorOperator
#
#******************************************************************************

def getFloor( n ):
    if isinstance( n, RPNMeasurement ):
        return RPNMeasurement( getFloor( n.value ), n.units )

    return floor( n )


@oneArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ) ] )
def getFloorOperator( n ):
    return getFloor( n )


#******************************************************************************
#
#  getCeilingOperator
#
#******************************************************************************

def getCeiling( n ):
    if isinstance( n, RPNMeasurement ):
        return RPNMeasurement( getCeiling( n.value ), n.units )

    return ceil( n )


@oneArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ) ] )
def getCeilingOperator( n ):
    return getCeiling( n )


#******************************************************************************
#
#  getMaximumOperator
#
#******************************************************************************

@listArgFunctionEvaluator( )
@argValidator( [ ListValidator( ) ] )
def getMaximumOperator( n ):
    if isinstance( n[ 0 ], ( list, RPNGenerator ) ):
        return [ max( arg ) for arg in n ]

    return max( n )


#******************************************************************************
#
#  getMinimumOperator
#
#******************************************************************************

@listArgFunctionEvaluator( )
@argValidator( [ ListValidator( ) ] )
def getMinimumOperator( n ):
    if isinstance( n[ 0 ], ( list, RPNGenerator ) ):
        return [ min( arg ) for arg in n ]

    return min( n )


#******************************************************************************
#
#  isEvenOperator
#
#******************************************************************************

def isEven( n ):
    return 1 if fmod( n, 2 ) == 0 else 0


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( ) ] )
def isEvenOperator( n ):
    return isEven( n )


#******************************************************************************
#
#  isOddOperator
#
#******************************************************************************

def isOdd( n ):
    return 1 if fmod( n, 2 ) == 1 else 0


@oneArgFunctionEvaluator( )
@argValidator( [ IntValidator( ) ] )
def isOddOperator( n ):
    return isOdd( n )


#******************************************************************************
#
#  isNotZeroOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ) ] )
def isNotZeroOperator( n ):
    return 0 if n == 0 else 1


#******************************************************************************
#
#  isZeroOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ) ] )
def isZeroOperator( n ):
    return 1 if n == 0 else 0


#******************************************************************************
#
#  getMantissaOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ RealValidator( ) ] )
def getMantissaOperator( n ):
    return fmod( fabs( n ), 1 )


#******************************************************************************
#
#  getModuloOperator
#
#******************************************************************************

@twoArgFunctionEvaluator( )
@argValidator( [ RealOrMeasurementValidator( ), RealOrMeasurementValidator( ) ] )
def getModuloOperator( n, k ):
    if isinstance( n, RPNMeasurement ):
        return n.getModulo( k )

    if isinstance( k, RPNMeasurement ):
        raise ValueError( 'cannot take a non-measurement modulo a measurement' )

    return fmod( n, k )


#******************************************************************************
#
#  getAGMOperator
#
#******************************************************************************

@twoArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ), MultiplicativeValidator( ) ] )
def getAGMOperator( n, k ):
    return agm( n, k )


#******************************************************************************
#
#  getExpOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ) ] )
def getExpOperator( n ):
    return exp( n )


#******************************************************************************
#
#  getExp10Operator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ) ] )
def getExp10Operator( n ):
    return power( 10, n )


#******************************************************************************
#
#  getExpPhiOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ) ] )
def getExpPhiOperator( n ):
    return power( phi, n )


#******************************************************************************
#
#  getArgumentOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ) ] )
def getArgumentOperator( n ):
    return arg( n )


#******************************************************************************
#
#  getConjugateOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ) ] )
def getConjugateOperator( n ):
    return conj( n )


#******************************************************************************
#
#  getImaginaryOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ) ] )
def getImaginaryOperator( n ):
    return im( n )


#******************************************************************************
#
#  getRealOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ) ] )
def getRealOperator( n ):
    return re( n )


#******************************************************************************
#
#  getLambertWOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ) ] )
def getLambertWOperator( n ):
    return lambertw( n )


#******************************************************************************
#
#  getLIOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ) ] )
def getLIOperator( n ):
    return li( n )


#******************************************************************************
#
#  getLogOperator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ) ] )
def getLogOperator( n ):
    return ln( n )


#******************************************************************************
#
#  getLog10Operator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ) ] )
def getLog10Operator( n ):
    return log10( n )


#******************************************************************************
#
#  getLog2Operator
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ) ] )
def getLog2Operator( n ):
    return log( n, 2 )


#******************************************************************************
#
#  getLogXYOperator
#
#******************************************************************************

@twoArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ), MultiplicativeValidator( ) ] )
def getLogXYOperator( n, k ):
    return log( n, k )


#******************************************************************************
#
#  getPolyexpOperator
#
#******************************************************************************

@twoArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ), MultiplicativeValidator( ) ] )
def getPolyexpOperator( n, k ):
    return polyexp( n, k )


#******************************************************************************
#
#  getPolylogOperator
#
#******************************************************************************

@twoArgFunctionEvaluator( )
@argValidator( [ MultiplicativeValidator( ), MultiplicativeValidator( ) ] )
def getPolylogOperator( n, k ):
    return polylog( n, k )


#******************************************************************************
#
#  calculateHypotenuseOperator
#
#******************************************************************************

@twoArgFunctionEvaluator( )
@argValidator( [ RealValidator( 0 ), RealValidator( 0 ) ] )
def calculateHypotenuseOperator( n, k ):
    return hypot( n, k )


#******************************************************************************
#
#  calculateNthHyperoperatorOperator
#
#******************************************************************************

class HyperopLeft( hyperop ):
    def __call__( self, a, b ):
        '''
        Evaluate and return expression H[n](a,b).
        (a,b) must be non-negative for n>4.
        '''
        check = self._check_value( a, b )

        if check is not None:
            return check

        # Apply fold
        #return reduce( lambda x, y: self.lower( x, y ), self._repeat( a, b ) )
        return reduce( self.lower, self._repeat( a, b ) )


@argValidator( [ IntValidator( 0 ), RealValidator( 0 ), IntValidator( 0 ) ] )
def calculateNthHyperoperatorOperator( a, b, c ):
    if a == 0:
        return c + 1

    if b == 0:
        if c == 1:
            return 0

        return 1

    if c == 0:
        return 1

    if a > 1 and b == 1:
        return 1

    if a > 1 and c == 1:
        return b

    if b == 2 and c == 2:
        if a == 0:
            return 3

        return 4

    if a > 4 and c > 2:
        raise ValueError( 'overflow' )

    if a > 4 and b > 4:
        raise ValueError( 'overflow' )

    return HyperopLeft( a )( b, c )


#******************************************************************************
#
#  calculateNthRightHyperoperatorOperator
#
#******************************************************************************

@argValidator( [ IntValidator( 0 ), RealValidator( 0 ), IntValidator( 0 ) ] )
def calculateNthRightHyperoperatorOperator( a, b, c ):
    if a == 0:
        return c + 1

    if b == 0:
        if c == 1:
            return 0

        return 1

    if c == 0:
        return 1

    if a > 1 and b == 1:
        return 1

    if a > 1 and c == 1:
        return b

    if b == 2 and c == 2:
        if a == 0:
            return 3

        return 4

    if a > 4 and c > 2:
        raise ValueError( 'overflow' )

    if a > 4 and b > 5:
        raise ValueError( 'overflow' )

    return hyperop( a )( b, c )


#******************************************************************************
#
#  getPowModOperator
#
#******************************************************************************

def getPowMod( a, b, c ):
    '''
    Calculate (a ** y) % z efficiently.
    '''
    result = 1

    while b:
        if fmod( b, 2 ) == 1:
            result = fmod( fmul( result, a ), c )

        b = floor( fdiv( b, 2 ) )
        a = fmod( fmul( a, a ), c )

    return result


@argValidator( [ IntValidator( ), IntValidator( ), IntValidator( 1 ) ] )
def getPowModOperator( a, b, c ):
    return pow( int( a ), int( b ), int( c ) )


#******************************************************************************
#
#  getPowModOperatorNew
#
#******************************************************************************

@argValidator( [ IntValidator( ), IntValidator( ), IntValidator( 1 ) ] )
def getPowModOperatorNew( a, b, c ):
    return getPowMod( a, b, c )





#!/usr/bin/env python

#******************************************************************************
#
#  rpnOperator.py
#
#  rpnChilada operator class definitions
#  copyright (c) 2021, Rick Gutleber (rickg@his.com)
#
#  License: GNU GPL 3.0 (see <http://www.gnu.org/licenses/gpl.html> for more
#  information).
#
#******************************************************************************

import random

from mpmath import exp, fadd, fmul, nan, nstr

from rpn.rpnGenerator import RPNGenerator
from rpn.rpnMeasurementClass import RPNMeasurement
from rpn.rpnSpecial import getRandomInteger, getRandomNumber
from rpn.rpnUtils import abortArgsNeeded
from rpn.rpnValidator import RPNValidator

import rpn.rpnGlobals as g


#******************************************************************************
#
#  generateDefaultArgument
#
#******************************************************************************

def generateDefaultArgument( ):
    return 'argument'


#******************************************************************************
#
#  generateRealArgument
#
#  for real numbers we are treating the range as logarithms
#
#******************************************************************************

def generateRealArgument( validRange=None, allowNegative=True ):
    if validRange is None:
        validRange = [ 0, 10 ]

    factor = 1

    if allowNegative and getRandomInteger( 2 ) == 1:
        factor = -1

    return nstr( fmul( exp( fmul( getRandomNumber( ), random.uniform( *validRange ) ) ), factor ) )


def generateIntegerArgument( validRange = None, allowNegative=True ):
    if validRange is None:
        validRange = [ 0, 1_000_000_000 ]

    factor = 1

    if allowNegative and getRandomInteger( 2 ) == 1:
        factor = -1

    return str( fmul( fadd( getRandomInteger( validRange[ 1 ] ), validRange[ 0 ] ), factor ) )


def generateStringArgument( ):
    return 'argument'


def generateDateTimeArgument( ):
    return 'argument'


def generateLocationArgument( ):
    return 'argument'


def generateBooleanArgument( ):
    return 'argument'


def generateMeasurementArgument( ):
    return 'argument'


def generateAstronomicalObjectArgument( ):
    return 'argument'


def generateListArgument( ):
    return 'argument'


def generateGeneratorArgument( ):
    return 'argument'


def generateFunctionArgument( ):
    return 'argument'


#******************************************************************************
#
#  argumentGenerators dict
#
#******************************************************************************

argumentGenerators = {
    RPNValidator.Default             : generateDefaultArgument,
    RPNValidator.Real                : generateRealArgument,
    RPNValidator.Integer             : generateIntegerArgument,
    RPNValidator.String              : generateStringArgument,
    RPNValidator.DateTime            : generateDateTimeArgument,
    RPNValidator.Location            : generateLocationArgument,
    RPNValidator.Boolean             : generateBooleanArgument,
    RPNValidator.Measurement         : generateMeasurementArgument,
    RPNValidator.AstronomicalObject  : generateAstronomicalObjectArgument,
    RPNValidator.List                : generateListArgument,
    RPNValidator.Generator           : generateGeneratorArgument,
    RPNValidator.Function            : generateFunctionArgument,
}


#******************************************************************************
#
#  class RPNVariable
#
#******************************************************************************

class RPNVariable( ):
    '''
    This class represents a variable in rpn, and it maintains a global
    dictionary of all variables keyed by name.
    '''
    def __init__( self, name ):
        self.name = name
        self.isHistory = not name[ 0 ].isalpha( )

        if name in g.variables:
            self.value = g.variables[ name ]
        else:
            self.value = nan

    def setValue( self, value ):
        if self.isHistory:
            raise ValueError( 'cannot set the value of a history expression' )

        self.value = value
        g.variables[ self.name ] = value

    def getValue( self ):
        if self.isHistory:
            prompt = int( self.name )

            if 0 < prompt < g.promptCount:
                self.value = g.results[ prompt ]
            else:
                raise ValueError( 'result index out of range' )

        return self.value


#******************************************************************************
#
#  class RPNOperator
#
#******************************************************************************

class RPNOperator( ):
    '''This class represents all the data needed to define an operator.'''
    def __init__( self, function, argCount ):
        self.function = function
        self.argCount = argCount

    def evaluate( self, term, index, currentValueList ):
        # handle a regular operator
        argsNeeded = self.argCount

        # first we validate, and make sure the operator has enough arguments
        if len( currentValueList ) < argsNeeded:
            abortArgsNeeded( term, index, argsNeeded )
            return False

        if argsNeeded == 0:
            result = self.function( )
        else:
            argList = list( )

            if g.operatorList:
                g.operatorsInList += 1

            # build argument list
            for i in range( 0, argsNeeded ):
                if g.operatorList:
                    # we need to tee a generator so it can run more than once
                    if isinstance( currentValueList[ g.lastOperand - 1 ], RPNGenerator ):
                        arg = currentValueList[ g.lastOperand - i ].clone( )
                    else:
                        arg = currentValueList[ g.lastOperand - i ]

                    if argsNeeded > g.operandsToRemove:
                        g.operandsToRemove = argsNeeded
                else:
                    arg = checkForVariable( currentValueList.pop( ) )

                    if term != 'set_variable' and isinstance( arg, RPNVariable ):
                        raise ValueError( 'set_variable called with a nonvariable argument' )
                        #arg = arg.getValue( )

                argList.append( arg if isinstance( arg, ( list, RPNGenerator ) ) else [ arg ] )

            # argument validation
            #self.validateArgTypes( term, *reverse( argList ), self.argTypes[ i ] )

            #print( 'argList', *reversed( argList ) )
            #print( 'self.function', self.function )

            result = callers[ argsNeeded ]( self.function, *reversed( argList ) )
            #result = list( map( self.function, *reversed( argList ) ) )

        # process results
        newResult = list( )

        if isinstance( result, RPNGenerator ):
            newResult.append( result )
        else:
            if not isinstance( result, list ):
                result = [ result ]

            for item in result:
                if isinstance( item, RPNMeasurement ) and item.units == { }:
                    newResult.append( item.value )
                else:
                    newResult.append( item )

        if len( newResult ) == 1:
            newResult = newResult[ 0 ]

        if term not in sideEffectOperators:
            currentValueList.append( newResult )

        return True

    #def generateCall( self, operatorName ):
    #    args = [ ]
    #
    #    print( operatorName )
    #    for i in range( self.argCount ):
    #        args.append( argumentGenerators[ self.argTypes[ i ] ]( ) )
    #
    #    if args:
    #        print( args )
    #        return ' '.join( args ) + ' ' + operatorName
    #
    #    return operatorName


#******************************************************************************
#
#  checkForVariable
#
#******************************************************************************

def checkForVariable( term ):
    if not isinstance( term, str ):
        return term

    # first check for a variable name or history expression
    if not term or term[ 0 ] != '$':
        return term

    if not g.interactive and term[ 1 : ] in g.userVariables:
        return g.userVariables[ term[ 1 : ] ]

    return RPNVariable( term[ 1 : ] )


#******************************************************************************
#
#  callers
#
#******************************************************************************

callers = [
    lambda func, args: func( ),
    lambda func, arg: func( arg ),
    lambda func, arg1, arg2: func( arg1, arg2 ),

    # 3, 4, and 5 argument functions don't recurse with lists more than one level
    # I have some ideas about how to improve this.
    lambda func, arg1, arg2, arg3:
        [ func( a, b, c ) for a in arg1 for b in arg2 for c in arg3 ],
    lambda func, arg1, arg2, arg3, arg4:
        [ func( a, b, c, d ) for a in arg1 for b in arg2 for c in arg3 for d in arg4 ],
    lambda func, arg1, arg2, arg3, arg4, arg5:
        [ func( a, b, c, d, e ) for a in arg1 for b in arg2 for c in arg3 for d in arg4 for e in arg5 ],
]


#******************************************************************************
#
#  sideEffectOperators
#
#  This is a list of operators that execute without modifying the result
#  stack.
#
#******************************************************************************

sideEffectOperators = [
    'comma_mode',
    'hex_mode',
    'identify_mode',
    'leading_zero_mode',
    'octal_mode',
    'timer_mode',
]



#!/usr/bin/env python

#******************************************************************************
#
#  rpnValidator.py
#
#  rpnChilada argument validation classes
#  copyright (c) 2021, Rick Gutleber (rickg@his.com)
#
#  License: GNU GPL 3.0 (see <http://www.gnu.org/licenses/gpl.html> for more
#  information).
#
#******************************************************************************

import functools

from mpmath import floor, im, mpf
from mpmath.ctx_mp_python import mpc

from rpn.rpnDateTimeClass import RPNDateTime
from rpn.rpnGenerator import RPNGenerator
from rpn.rpnLocationClass import RPNLocation
from rpn.rpnMeasurementClass import RPNMeasurement


#******************************************************************************
#
#  argValidator
#
#******************************************************************************

def argValidator( validators ):
    def argValidatorFunction( func ):
        @functools.wraps( func )
        def validateArgs( *args ):
            newArgs = [ ]
            for index, validator in enumerate( validators ):
                newArgs.append( validator.validate( args[ index ] ) )

            return func( *newArgs )

        return validateArgs

    return argValidatorFunction


#******************************************************************************
#
#  RPNValidator class
#
#******************************************************************************

class RPNValidator( ):
    '''
    This is a class used to validate arguments for operators.
    '''
    Default =               0           # any argument is valid
    Real =                  1
    Integer =               1 << 1
    Complex =               1 << 2
    ComplexInteger =        1 << 3
    String =                1 << 4
    DateTime =              1 << 5
    Location =              1 << 6      # location object (operators will automatically convert a string)
    Boolean =               1 << 7      # 0 or 1
    Measurement =           1 << 8
    Length =                1 << 9
    AstronomicalObject =    1 << 10
    List =                  1 << 11     # the argument must be a list
    Generator =             1 << 12     # Generator is a separate type, but eventually it should be equivalent to List
    Function =              1 << 13
    TimeZone =              1 << 14
    Year =                  1 << 15
    Comparable =            1 << 16     # real, or measurement, or date-time
    Additive =              1 << 17     # a value that can have something added to or subtracted from
    Multiplicative =        1 << 18     # a value that can be multiplied (complex, measurement)

    def __init__( self, valueType=Default, minValue=None, maxValue=None, specials=None ):
        self.valueType = valueType
        self.minValue = minValue
        self.maxValue = maxValue
        self.specials = specials

    def validate( self, argument ):
        if self.valueType == self.Default:
            pass
        elif self.valueType == self.Integer:
            argument = self.validateInt( argument )
        elif self.valueType == self.Real:
            argument = self.validateReal( argument )
        elif self.valueType == self.Complex:
            argument = self.validateComplex( argument )
        elif self.valueType == self.Integer + self.Measurement:
            argument = self.validateIntOrMeasurement( argument )
        elif self.valueType == self.Real + self.Measurement:
            argument = self.validateRealOrMeasurement( argument )
        elif self.valueType == self.Complex + self.Measurement:
            argument = self.validateComplexOrMeasurement( argument )
        elif self.valueType == self.Real + self.Measurement + self.DateTime:
            argument = self.validateComparable( argument )
        elif self.valueType == self.Comparable:
            argument = self.validateAdditive( argument )
        elif self.valueType == self.Length:
            argument = self.validateLength( argument )
        elif self.valueType == self.List:
            argument = self.validateList( argument )
        elif self.valueType == self.Integer + self.String + self.Measurement:
            argument = self.validateElement( argument )
        elif self.valueType == self.Location:
            argument = self.validateLocation( argument )
        elif self.valueType == self.Location + self.DateTime:
            argument = self.validateLocationOrDateTime( argument )         
        elif self.valueType == self.Year:
            argument = self.validateYear( argument )

        if self.specials:
            for special in self.specials:
                if not special[ 0 ]( argument ):
                    raise ValueError( special[ 1 ] )

        return argument

    def addSpecial( self, func, formatString ):
        if self.specials:
            self.specials.append( ( func, formatString ) )
        else:
            self.specials = [ ( func, formatString ) ]

    def validateInt( self, argument ):
        if not isinstance( argument, ( complex, mpc, mpf, int, float ) ):
            raise ValueError( f'\'type\' { type( argument ) } found, integer value expected' )

        if im( argument ) != 0:
            raise ValueError( 'real argument expected ({})'.format( argument ) )

        if argument != floor( argument ):
            raise ValueError( 'integer argument expected ({})'.format( argument ) )

        if self.minValue is not None and argument < self.minValue:
            raise ValueError( f'argument value is { int( argument ) }, but the minimum valid value is { int( self.minValue ) }.' )

        if self.maxValue is not None and argument > self.maxValue:
            raise ValueError( f'argument value is { int( argument ) }, but the maximum valid value is { int( self.maxValue ) }.' )

        return argument

    def validateReal( self, argument ):
        if not isinstance( argument, ( complex, mpc, mpf, int, float ) ):
            raise ValueError( f'\'type\' { type( argument ) } found, numeric value expected' )

        if im( argument ) != 0:
            raise ValueError( 'real argument expected ({})'.format( argument ) )

        if self.minValue is not None and argument < self.minValue:
            raise ValueError( f'argument value is { argument }, minimum valid value is { self.minValue }.' )

        if self.maxValue is not None and argument > self.maxValue:
            raise ValueError( f'argument value is { argument }, maximum valid value is { self.maxValue }.' )

        return argument

    def validateComplex( self, argument ):
        if not isinstance( argument, ( complex, mpc, mpf, int, float ) ):
            raise ValueError( f'\'type\' { type( argument ) } found, numeric value expected' )

        if self.minValue is not None :
            raise ValueError( 'The min constraint is invalid for validating complex arguments.' )

        if self.maxValue is not None :
            raise ValueError( 'The max constraint is invalid for validating complex arguments.' )

        return argument

    def validateDateTime( self, argument ):
        if not isinstance( argument, RPNDateTime ):
            raise ValueError( f'argument is type { type( argument ) }, but date-time value expected' )

        return argument

    def validateString( self, argument ):
        if not isinstance( argument, str ):
            raise ValueError( f'argument is type { type( argument ) }, but string value expected' )

        return argument

    def validateMeasurement( self, argument ):
        if not isinstance( argument, RPNMeasurement ):
            raise ValueError( f'argument is type { type( argument ) }, but measurement value expected' )

        return argument

    def validateIntOrMeasurement( self, argument ):
        if isinstance( argument, ( complex, mpc, mpf, int, float ) ):
            argument = self.validateInt( argument )
        elif isinstance( argument, RPNMeasurement ):
            self.validateMeasurement( argument )
        else:
            raise ValueError( f'\'type\' { type( argument ) } found, integer or measurement value expected' )

        return argument

    def validateRealOrMeasurement( self, argument ):
        if isinstance( argument, ( complex, mpc, mpf, int, float ) ):
            argument = self.validateReal( argument )
        elif isinstance( argument, RPNMeasurement ):
            pass
        else:
            raise ValueError( f'argument is type { type( argument ) }, but numeric or measurement value expected' )

        return argument

    def validateComplexOrMeasurement( self, argument ):
        if isinstance( argument, ( complex, mpc, mpf, int, float ) ):
            argument = self.validateComplex( argument )
        elif isinstance( argument, RPNMeasurement ):
            pass
        else:
            raise ValueError( f'\'type\' { type( argument ) } found, numeric or measurement value expected' )

        return argument

    def validateComparable( self, argument ):
        if isinstance( argument, ( complex, mpc, mpf, int, float ) ):
            argument = self.validateReal( argument )
        elif isinstance( argument, ( RPNMeasurement, RPNDateTime ) ):
            pass
        else:
            raise ValueError( f'\'type\' { type( argument ) } found, numeric, measurement, date-time value expected' )

        return argument

    def validateAdditive( self, argument ):
        if isinstance( argument, ( complex, mpc, mpf, int, float ) ):
            argument = self.validateComplex( argument )
        elif isinstance( argument, ( RPNMeasurement, RPNDateTime ) ):
            pass
        else:
            raise ValueError( f'\'type\' { type( argument ) } found, numeric, measurement, date-time value expected' )

        return argument

    def validateLength( self, argument ):
        if isinstance( argument, ( complex, mpc, mpf, int, float ) ):
            self.validateReal( argument )
            argument = RPNMeasurement( argument, 'meter' )
        elif isinstance( argument, RPNMeasurement ):
            if argument.getDimensions( ) != { 'length' : 1 }:
                raise ValueError( 'measurement argument must be a length' )
        else:
            raise ValueError( f'\'type\' { type( argument ) } found, measurement (length) expected' )

        return argument

    def validateList( self, argument ):
        if not isinstance( argument, ( list, RPNGenerator ) ):
            raise ValueError( f'\'type\' { type( argument ) } found, list expected' )

        return argument

    def validateElement( self, argument ):
        if isinstance( argument, ( complex, mpc, mpf, int, float ) ):
            self.validateInt( argument )
        elif isinstance( argument, RPNMeasurement ):
            self.validateMeasurement( argument )
        elif not isinstance( argument, str ):
            raise ValueError( f'\'type\' { type( argument ) } found, string or integer expected' )

        return argument

    def validateLocation( self, argument ):
        if not isinstance( argument, ( str, RPNLocation ) ):
            raise ValueError( f'\'type\' { type( argument ) } found, string or location object expected' )

        return argument

    def validateLocationOrDateTime( self, argument ):
        if not isinstance( argument, ( str, RPNLocation, RPNDateTime ) ):
            raise ValueError( f'\'type\' { type( argument ) } found, string, date-time or location object expected' )

        return argument

    def validateYear( self, argument ):
        if isinstance( argument, ( complex, mpc, mpf, int, float ) ):
            argument = self.validateInt( argument )
        elif isinstance( argument, RPNDateTime ):
            self.validateDateTime( argument )
            argument = argument.year
        else:
            raise ValueError( f'\'type\' { type( argument ) } found, integer or date-time value expected' )

        return argument


class DefaultValidator( RPNValidator ):
    def __init__( self, specials=None ):
        super( ).__init__( RPNValidator.Default, specials=specials )


class IntValidator( RPNValidator ):
    def __init__( self, minValue=None, maxValue=None, specials=None ):
        super( ).__init__( RPNValidator.Integer, minValue, maxValue, specials )


class RealValidator( RPNValidator ):
    def __init__( self, minValue=None, maxValue=None, specials=None ):
        super( ).__init__( RPNValidator.Real, minValue, maxValue, specials )


class ComplexValidator( RPNValidator ):
    def __init__( self, minValue=None, maxValue=None, specials=None ):
        super( ).__init__( RPNValidator.Complex, minValue, maxValue, specials )


class MeasurementValidator( RPNValidator ):
    def __init__( self, minValue=None, maxValue=None, specials=None ):
        super( ).__init__( RPNValidator.Measurement, minValue, maxValue, specials=specials )


class StringValidator( RPNValidator ):
    def __init__( self, specials=None ):
        super( ).__init__( RPNValidator.String, specials )


class LengthValidator( RPNValidator ):
    def __init__( self, minValue=None, maxValue=None, specials=None ):
        super( ).__init__( RPNValidator.Length, minValue, maxValue, specials=specials )


class DateTimeValidator( RPNValidator ):
    def __init__( self, minValue=None, maxValue=None, specials=None ):
        super( ).__init__( RPNValidator.DateTime, minValue, maxValue, specials=specials )


class ListValidator( RPNValidator ):
    def __init__( self, specials=None ):
        super( ).__init__( RPNValidator.List, specials=specials )


class YearValidator( RPNValidator ):
    def __init__( self, specials=None ):
        super( ).__init__( RPNValidator.Year, specials=specials )


class LocationValidator( RPNValidator ):
    def __init__( self, specials=None ):
        super( ).__init__( RPNValidator.Location, specials=specials )


class ComparableValidator( RPNValidator ):
    def __init__( self, specials=None ):
        super( ).__init__( RPNValidator.Comparable, specials=specials )


class AdditiveValidator( RPNValidator ):
    def __init__( self, specials=None ):
        super( ).__init__( RPNValidator.Additive, specials=specials )


class MultiplicativeValidator( RPNValidator ):
    def __init__( self, specials=None ):
        super( ).__init__( RPNValidator.Multiplicative, specials=specials )


class ElementValidator( RPNValidator ):
    def __init__( self, specials=None ):
        super( ).__init__( RPNValidator.Integer + RPNValidator.String, specials=specials )


# compound validators

class IntOrMeasurementValidator( RPNValidator ):
    def __init__( self, specials=None ):
        super( ).__init__( RPNValidator.Integer + RPNValidator.Measurement, specials=specials )


class RealOrMeasurementValidator( RPNValidator ):
    def __init__( self, specials=None ):
        super( ).__init__( RPNValidator.Real + RPNValidator.Measurement, specials=specials )


class LocationOrDateTimeValidator( RPNValidator ):
    def __init__( self, specials=None ):
        super( ).__init__( RPNValidator.Location + RPNValidator.DateTime, specials=specials )
